@@ -75,6 +75,7 @@ struct sms_data {
 	gboolean cnma_enabled;
 	char *cnma_ack_pdu;
 	int cnma_ack_pdu_len;
+	guint timeout_source;
 	GAtChat *chat;
 	unsigned int vendor;
 };
@@ -91,7 +92,6 @@ static void at_csca_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	ofono_sms_sca_set_cb_t cb = cbd->cb;
 	struct ofono_error error;
 
-	dump_response("csca_set_cb", ok, result);
 	decode_at_error(&error, g_at_result_final_response(result));
 
 	cb(&error, cbd->data);
@@ -108,7 +108,7 @@ static void at_csca_set(struct ofono_sms *sms,
 	if (!cbd)
 		goto error;
 
-	sprintf(buf, "AT+CSCA=\"%s\",%d", sca->number, sca->type);
+	snprintf(buf, sizeof(buf), "AT+CSCA=\"%s\",%d", sca->number, sca->type);
 
 	if (g_at_chat_send(data->chat, buf, csca_prefix,
 				at_csca_set_cb, cbd, g_free) > 0)
@@ -130,7 +130,6 @@ static void at_csca_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_phone_number sca;
 	const char *number;
 
-	dump_response("at_csca_cb", ok, result);
 	decode_at_error(&error, g_at_result_final_response(result));
 
 	if (!ok) {
@@ -158,7 +157,7 @@ static void at_csca_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 	g_at_result_iter_next_number(&iter, &sca.type);
 
-	ofono_debug("csca_query_cb: %s, %d", sca.number, sca.type);
+	DBG("csca_query_cb: %s, %d", sca.number, sca.type);
 
 	cb(&error, &sca, cbd->data);
 
@@ -196,7 +195,6 @@ static void at_cmgs_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_error error;
 	int mr;
 
-	dump_response("at_cmgs_cb", ok, result);
 	decode_at_error(&error, g_at_result_final_response(result));
 
 	if (!ok) {
@@ -212,7 +210,7 @@ static void at_cmgs_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	if (!g_at_result_iter_next_number(&iter, &mr))
 		goto err;
 
-	ofono_debug("Got MR: %d", mr);
+	DBG("Got MR: %d", mr);
 
 	cb(&error, mr, cbd->data);
 	return;
@@ -234,12 +232,12 @@ static void at_cmgs(struct ofono_sms *sms, unsigned char *pdu, int pdu_len,
 		goto error;
 
 	if (mms) {
-		sprintf(buf, "AT+CMMS=%d", mms);
+		snprintf(buf, sizeof(buf), "AT+CMMS=%d", mms);
 		g_at_chat_send(data->chat, buf, none_prefix,
 				NULL, NULL, NULL);
 	}
 
-	len = sprintf(buf, "AT+CMGS=%d\r", tpdu_len);
+	len = snprintf(buf, sizeof(buf), "AT+CMGS=%d\r", tpdu_len);
 	encode_hex_own_buf(pdu, pdu_len, 0, buf+len);
 
 	if (g_at_chat_send(data->chat, buf, cmgs_prefix,
@@ -289,21 +287,19 @@ static void at_cds_notify(GAtResult *result, gpointer user_data)
 	const char *pdu;
 	char buf[256];
 
-	dump_response("at_cds_notify", TRUE, result);
-
 	if (!at_parse_pdu_common(result, "+CDS:", &pdu, &pdulen)) {
 		ofono_error("Unable to parse CDS notification");
 		return;
 	}
 
-	ofono_debug("Got new Status-Report PDU via CDS: %s, %d", pdu, pdulen);
+	DBG("Got new Status-Report PDU via CDS: %s, %d", pdu, pdulen);
 
 	/* We must acknowledge the PDU using CNMA */
 	if (data->cnma_ack_pdu)
-		sprintf(buf, "AT+CNMA=1,%d\r%s", data->cnma_ack_pdu_len,
-				data->cnma_ack_pdu);
-	else
-		sprintf(buf, "AT+CNMA=0"); /* Should be a safe fallback */
+		snprintf(buf, sizeof(buf), "AT+CNMA=1,%d\r%s",
+				data->cnma_ack_pdu_len, data->cnma_ack_pdu);
+	else /* Should be a safe fallback */
+		snprintf(buf, sizeof(buf), "AT+CNMA=0");
 
 	g_at_chat_send(data->chat, buf, none_prefix, at_cnma_cb, NULL, NULL);
 }
@@ -318,8 +314,6 @@ static void at_cmt_notify(GAtResult *result, gpointer user_data)
 	unsigned char pdu[176];
 	char buf[256];
 
-	dump_response("at_cmt_notify", TRUE, result);
-
 	if (!at_parse_pdu_common(result, "+CMT:", &hexpdu, &tpdu_len)) {
 		ofono_error("Unable to parse CMT notification");
 		return;
@@ -330,18 +324,17 @@ static void at_cmt_notify(GAtResult *result, gpointer user_data)
 		return;
 	}
 
-	ofono_debug("Got new SMS Deliver PDU via CMT: %s, %d",
-			hexpdu, tpdu_len);
+	DBG("Got new SMS Deliver PDU via CMT: %s, %d", hexpdu, tpdu_len);
 
 	decode_hex_own_buf(hexpdu, -1, &pdu_len, 0, pdu);
 	ofono_sms_deliver_notify(sms, pdu, pdu_len, tpdu_len);
 
 	/* We must acknowledge the PDU using CNMA */
 	if (data->cnma_ack_pdu)
-		sprintf(buf, "AT+CNMA=1,%d\r%s", data->cnma_ack_pdu_len,
-				data->cnma_ack_pdu);
-	else
-		sprintf(buf, "AT+CNMA=0"); /* Should be a safe fallback */
+		snprintf(buf, sizeof(buf), "AT+CNMA=1,%d\r%s",
+				data->cnma_ack_pdu_len, data->cnma_ack_pdu);
+	else /* Should be a safe fallback */
+		snprintf(buf, sizeof(buf), "AT+CNMA=0");
 
 	g_at_chat_send(data->chat, buf, none_prefix, at_cnma_cb, NULL, NULL);
 }
@@ -355,8 +348,6 @@ static void at_cmgr_notify(GAtResult *result, gpointer user_data)
 	long pdu_len;
 	int tpdu_len;
 
-	dump_response("at_cmgr_notify", TRUE, result);
-
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "+CMGR:"))
@@ -376,7 +367,7 @@ static void at_cmgr_notify(GAtResult *result, gpointer user_data)
 	if (strlen(hexpdu) > sizeof(pdu) * 2)
 		goto err;
 
-	ofono_debug("Got PDU: %s, with len: %d", hexpdu, tpdu_len);
+	DBG("Got PDU: %s, with len: %d", hexpdu, tpdu_len);
 
 	decode_hex_own_buf(hexpdu, -1, &pdu_len, 0, pdu);
 	ofono_sms_deliver_notify(sms, pdu, pdu_len, tpdu_len);
@@ -412,11 +403,11 @@ static void at_cmti_cpms_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 	data->store = req->store;
 
-	sprintf(buf, "AT+CMGR=%d", req->index);
+	snprintf(buf, sizeof(buf), "AT+CMGR=%d", req->index);
 	g_at_chat_send(data->chat, buf, none_prefix, at_cmgr_cb, NULL, NULL);
 
 	/* We don't buffer SMS on the SIM/ME, send along a CMGD as well */
-	sprintf(buf, "AT+CMGD=%d", req->index);
+	snprintf(buf, sizeof(buf), "AT+CMGD=%d", req->index);
 	g_at_chat_send(data->chat, buf, none_prefix, at_cmgd_cb, NULL, NULL);
 }
 
@@ -429,8 +420,6 @@ static void at_cmti_notify(GAtResult *result, gpointer user_data)
 	GAtResultIter iter;
 	int index;
 
-	dump_response("at_cmti_notify", TRUE, result);
-
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "+CMTI:"))
@@ -449,7 +438,7 @@ static void at_cmti_notify(GAtResult *result, gpointer user_data)
 	if (!g_at_result_iter_next_number(&iter, &index))
 		goto err;
 
-	ofono_debug("Got a CMTI indication at %s, index: %d", strstore, index);
+	DBG("Got a CMTI indication at %s, index: %d", strstore, index);
 
 	if (store == data->store) {
 		struct cpms_request req;
@@ -468,8 +457,8 @@ static void at_cmti_notify(GAtResult *result, gpointer user_data)
 		req->store = store;
 		req->index = index;
 
-		sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"",
-			strstore, strstore, incoming);
+		snprintf(buf, sizeof(buf), "AT+CPMS=\"%s\",\"%s\",\"%s\"",
+				strstore, strstore, incoming);
 
 		g_at_chat_send(data->chat, buf, cpms_prefix, at_cmti_cpms_cb,
 				req, g_free);
@@ -502,8 +491,6 @@ static void at_cmgl_notify(GAtResult *result, gpointer user_data)
 	int status;
 	char buf[16];
 
-	dump_response("at_cmgl_notify", TRUE, result);
-
 	g_at_result_iter_init(&iter, result);
 
 	while (g_at_result_iter_next(&iter, "+CMGL:")) {
@@ -525,7 +512,7 @@ static void at_cmgl_notify(GAtResult *result, gpointer user_data)
 
 		hexpdu = g_at_result_pdu(result);
 
-		ofono_debug("Found an old SMS PDU: %s, with len: %d",
+		DBG("Found an old SMS PDU: %s, with len: %d",
 				hexpdu, tpdu_len);
 
 		if (strlen(hexpdu) > sizeof(pdu) * 2)
@@ -535,7 +522,7 @@ static void at_cmgl_notify(GAtResult *result, gpointer user_data)
 		ofono_sms_deliver_notify(sms, pdu, pdu_len, tpdu_len);
 
 		/* We don't buffer SMS on the SIM/ME, send along a CMGD */
-		sprintf(buf, "AT+CMGD=%d", index);
+		snprintf(buf, sizeof(buf), "AT+CMGD=%d", index);
 		g_at_chat_send(data->chat, buf, none_prefix,
 				at_cmgd_cb, NULL, NULL);
 	}
@@ -550,7 +537,7 @@ static void at_cmgl_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_sms *sms = user_data;
 
 	if (!ok)
-		ofono_debug("Initial listing SMS storage failed!");
+		DBG("Initial listing SMS storage failed!");
 
 	at_cmgl_done(sms);
 }
@@ -593,8 +580,8 @@ static void at_cmgl_set_cpms(struct ofono_sms *sms, int store)
 		req->sms = sms;
 		req->store = store;
 
-		sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"",
-			readwrite, readwrite, incoming);
+		snprintf(buf, sizeof(buf), "AT+CPMS=\"%s\",\"%s\",\"%s\"",
+				readwrite, readwrite, incoming);
 
 		g_at_chat_send(data->chat, buf, cpms_prefix, at_cmgl_cpms_cb,
 				req, g_free);
@@ -754,8 +741,6 @@ static void at_cnmi_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	gboolean supported = FALSE;
 	char buf[128];
 
-	dump_response("at_cnmi_query_cb", ok, result);
-
 	if (!ok)
 		goto out;
 
@@ -808,8 +793,6 @@ static void at_cpms_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_sms *sms = user_data;
 	struct sms_data *data = ofono_sms_get_data(sms);
 
-	dump_response("at_cpms_set_cb", ok, result);
-
 	if (ok)
 		return at_query_cnmi(sms);
 
@@ -820,7 +803,7 @@ static void at_cpms_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		return at_sms_not_supported(sms);
 	}
 
-	g_timeout_add_seconds(1, set_cpms, sms);
+	data->timeout_source = g_timeout_add_seconds(1, set_cpms, sms);
 }
 
 static gboolean set_cpms(gpointer user_data)
@@ -831,10 +814,14 @@ static gboolean set_cpms(gpointer user_data)
 	const char *incoming = storages[data->incoming];
 	char buf[128];
 
-	sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"", store, store, incoming);
+	snprintf(buf, sizeof(buf), "AT+CPMS=\"%s\",\"%s\",\"%s\"",
+			store, store, incoming);
 
 	g_at_chat_send(data->chat, buf, cpms_prefix,
 			at_cpms_set_cb, sms, NULL);
+
+	data->timeout_source = 0;
+
 	return FALSE;
 }
 
@@ -843,8 +830,6 @@ static void at_cmgf_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_sms *sms = user_data;
 	struct sms_data *data = ofono_sms_get_data(sms);
 
-	dump_response("at_cmgf_set_cb", ok, result);
-
 	if (ok) {
 		data->retries = 0;
 		set_cpms(sms);
@@ -854,11 +839,11 @@ static void at_cmgf_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	data->retries += 1;
 
 	if (data->retries == MAX_CMGF_RETRIES) {
-		ofono_debug("Unable to enter PDU mode");
+		DBG("Unable to enter PDU mode");
 		return at_sms_not_supported(sms);
 	}
 
-	g_timeout_add_seconds(1, set_cmgf, sms);
+	data->timeout_source = g_timeout_add_seconds(1, set_cmgf, sms);
 }
 
 static gboolean set_cmgf(gpointer user_data)
@@ -868,6 +853,9 @@ static gboolean set_cmgf(gpointer user_data)
 
 	g_at_chat_send(data->chat, "AT+CMGF=0", cmgf_prefix,
 			at_cmgf_set_cb, sms, NULL);
+
+	data->timeout_source = 0;
+
 	return FALSE;
 }
 
@@ -878,8 +866,6 @@ static void at_cpms_query_cb(gboolean ok, GAtResult *result,
 	struct sms_data *data = ofono_sms_get_data(sms);
 	gboolean supported = FALSE;
 
-	dump_response("cpms_query_cb", ok, result);
-
 	if (ok) {
 		int mem = 0;
 		GAtResultIter iter;
@@ -954,8 +940,6 @@ static void at_cmgf_query_cb(gboolean ok, GAtResult *result,
 	struct sms_data *data = ofono_sms_get_data(sms);
 	gboolean supported = FALSE;
 
-	dump_response("cmgf_query_cb", ok, result);
-
 	if (ok) {
 		GAtResultIter iter;
 		int mode;
@@ -989,8 +973,6 @@ static void at_csms_status_cb(gboolean ok, GAtResult *result,
 	struct sms_data *data = ofono_sms_get_data(sms);
 	gboolean supported = FALSE;
 
-	dump_response("csms_status_cb", ok, result);
-
 	if (ok) {
 		GAtResultIter iter;
 		int service;
@@ -1047,8 +1029,6 @@ static void at_csms_query_cb(gboolean ok, GAtResult *result,
 	int status_min, status_max;
 	char buf[128];
 
-	dump_response("csms_query_cb", ok, result);
-
 	if (!ok)
 		return at_sms_not_supported(sms);
 
@@ -1064,10 +1044,10 @@ static void at_csms_query_cb(gboolean ok, GAtResult *result,
 		if (status_min <= 1 && 1 <= status_max)
 			cnma_supported = TRUE;
 
-	ofono_debug("CSMS query parsed successfully");
+	DBG("CSMS query parsed successfully");
 
 out:
-	sprintf(buf, "AT+CSMS=%d", cnma_supported ? 1 : 0);
+	snprintf(buf, sizeof(buf), "AT+CSMS=%d", cnma_supported ? 1 : 0);
 	g_at_chat_send(data->chat, buf, csms_prefix,
 			at_csms_set_cb, sms, NULL);
 }
@@ -1097,6 +1077,9 @@ static void at_sms_remove(struct ofono_sms *sms)
 	if (data->cnma_ack_pdu)
 		g_free(data->cnma_ack_pdu);
 
+	if (data->timeout_source > 0)
+		g_source_remove(data->timeout_source);
+
 	g_free(data);
 }
 
