@@ -39,6 +39,8 @@
 #include <gatppp.h>
 
 static const char *none_prefix[] = { NULL };
+static const char *cfun_prefix[] = { "+CFUN:", NULL };
+static const char *creg_prefix[] = { "+CREG:", NULL };
 static const char *cgreg_prefix[] = { "+CGREG:", NULL };
 
 static gchar *option_ip = NULL;
@@ -49,9 +51,9 @@ static gint option_cid = 0;
 static gchar *option_apn = NULL;
 static gint option_offmode = 0;
 static gboolean option_legacy = FALSE;
-static gboolean option_ppp = FALSE;
 static gchar *option_username = NULL;
 static gchar *option_password = NULL;
+static gchar *option_pppdump = NULL;
 
 static GAtPPP *ppp;
 static GAtChat *control;
@@ -61,14 +63,16 @@ static GMainLoop *event_loop;
 enum state {
 	STATE_NONE = 0,
 	STATE_REGISTERING,
+	STATE_ATTACHING,
 	STATE_ACTIVATING
 };
 
 static int state = 0;
+static int oldmode = 0;
 
 static void gsmdial_debug(const char *str, void *data)
 {
-	g_print("%s: %s\n", (const char *)data, str);
+	g_print("%s: %s\n", (const char *) data, str);
 }
 
 static gboolean quit_eventloop(gpointer user_data)
@@ -100,12 +104,15 @@ static gboolean signal_cb(GIOChannel *channel, GIOCondition cond, gpointer data)
 	case SIGINT:
 	case SIGTERM:
 		if (terminated == 0) {
-			char buf[64];
-
 			g_timeout_add_seconds(10, quit_eventloop, NULL);
-			sprintf(buf, "AT+CFUN=%u", option_offmode);
-			g_at_chat_send(control, buf, none_prefix,
-					power_down, NULL, NULL);
+
+			if (ppp == NULL) {
+				char buf[64];
+				sprintf(buf, "AT+CFUN=%u", option_offmode);
+				g_at_chat_send(control, buf, none_prefix,
+						power_down, NULL, NULL);
+			} else
+				g_at_ppp_shutdown(ppp);
 		}
 
 		terminated++;
@@ -214,81 +221,57 @@ out:
 	return FALSE;
 }
 
-static void at_cgact_up_cb(gboolean ok, GAtResult *result, gpointer user_data)
-{
-	char buf[64];
-
-	if (!ok) {
-		g_print("Error activating context\n");
-		exit(1);
-	}
-
-	sprintf(buf, "AT+CGDATA=\"PPP\",%u", option_cid);
-
-	g_at_chat_send(modem, buf, none_prefix, NULL, NULL, NULL);
-}
-
-static void print_ip_address(const char *label, guint32 ip_addr)
-{
-	struct in_addr addr;
-	char buf[INET_ADDRSTRLEN];
-
-	addr.s_addr = ip_addr;
-
-	if (inet_ntop(AF_INET, &addr, buf, INET_ADDRSTRLEN))
-		g_print("%s: %s\n", label, buf);
-}
-
-static void ppp_connect(GAtPPP *ppp, GAtPPPConnectStatus success,
-			guint32 ip_addr, guint32 dns1, guint32 dns2,
+static void ppp_connect(const char *iface, const char *ip,
+			const char *dns1, const char *dns2,
 			gpointer user_data)
 {
-	if (success != G_AT_PPP_CONNECT_SUCCESS) {
-		g_print("Failed to create PPP interface!\n");
-		return;
-	}
-
 	/* print out the negotiated address and dns server */
-	print_ip_address("IP Address", ip_addr);
-	print_ip_address("Primary DNS Server", dns1);
-	print_ip_address("Secondary DNS Server", dns2);
+	g_print("Network Device: %s\n", iface);
+	g_print("IP Address: %s\n", ip);
+	g_print("Primary DNS Server: %s\n", dns1);
+	g_print("Secondary DNS Server: %s\n", dns2);
 }
 
-static void ppp_disconnect(GAtPPP *ppp, gpointer user_data)
+static void ppp_disconnect(GAtPPPDisconnectReason reason, gpointer user_data)
 {
-	g_print("PPP Link down\n");
+	char buf[64];
+	g_print("PPP Link down: %d\n", reason);
+	g_at_chat_resume(modem);
+
+	sprintf(buf, "AT+CFUN=%u", option_offmode);
+	g_at_chat_send(control, buf, none_prefix, power_down, NULL, NULL);
 }
 
 static void connect_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
-	GIOChannel *channel;
+	GAtIO *io;
 
 	if (!ok) {
 		g_print("Unable to define context\n");
 		exit(1);
 	}
 
-	if (option_ppp == FALSE)
-		return;
-
 	/* get the data IO channel */
-	channel = g_at_chat_get_channel(modem);
+	io = g_at_chat_get_io(modem);
 
 	/*
 	 * shutdown gatchat or else it tries to take all the input
 	 * from the modem and does not let PPP get it.
 	 */
-	g_at_chat_shutdown(control);
-	g_at_chat_shutdown(modem);
+	g_at_chat_suspend(modem);
 
 	/* open ppp */
-	ppp = g_at_ppp_new(channel);
+	ppp = g_at_ppp_new_from_io(io);
 	if (!ppp) {
 		g_print("Unable to create PPP object\n");
-		return;
+		exit(1);
 	}
-	g_at_ppp_set_credentials(ppp, option_username,
-				option_password);
+	g_at_ppp_set_debug(ppp, gsmdial_debug, "PPP");
+
+	if (option_pppdump)
+		g_at_ppp_set_recording(ppp, option_pppdump);
+
+	g_at_ppp_set_credentials(ppp, option_username, option_password);
 
 	/* set connect and disconnect callbacks */
 	g_at_ppp_set_connect_function(ppp, ppp_connect, NULL);
@@ -307,15 +290,12 @@ static void at_cgdcont_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		exit(1);
 	}
 
-	if (option_legacy == TRUE) {
+	if (option_legacy == TRUE)
 		sprintf(buf, "ATD*99***%u#", option_cid);
-		g_at_chat_send(modem, buf, none_prefix,
-				connect_cb, NULL, NULL);
-	} else {
-		sprintf(buf, "AT+CGACT=1,%u", option_cid);
-		g_at_chat_send(control, buf, none_prefix,
-				at_cgact_up_cb, NULL, NULL);
-	}
+	else
+		sprintf(buf, "AT+CGDATA=\"PPP\",%u", option_cid);
+
+	g_at_chat_send(modem, buf, none_prefix, connect_cb, NULL, NULL);
 }
 
 static void setup_context(int status)
@@ -326,7 +306,7 @@ static void setup_context(int status)
 	state = STATE_ACTIVATING;
 
 	g_print("Registered to GPRS network, roaming=%s\n",
-			status == 5 ? "True" : "False");
+					status == 5 ? "true" : "false");
 
 	len = sprintf(buf, "AT+CGDCONT=%u,\"IP\"", option_cid);
 	snprintf(buf + len, sizeof(buf) - len - 3, ",\"%s\"", option_apn);
@@ -337,11 +317,11 @@ static void cgreg_notify(GAtResult *result, gpointer user_data)
 {
 	int status, lac, ci, tech;
 
-	if (state != STATE_REGISTERING)
+	if (state != STATE_ATTACHING)
 		return;
 
 	if (at_util_parse_reg_unsolicited(result, "+CGREG:", &status,
-				&lac, &ci, &tech) == FALSE)
+						&lac, &ci, &tech) == FALSE)
 		return;
 
 	if (status != 1 && status != 5)
@@ -354,13 +334,19 @@ static void cgreg_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	int status, lac, ci, tech;
 
-	if (at_util_parse_reg(result, "+CGREG:", NULL, &status,
-				&lac, &ci, &tech) == FALSE)
+	if (!ok)
 		return;
 
-	if (status != 1 && status != 5)
+	if (at_util_parse_reg(result, "+CGREG:", NULL, &status,
+						&lac, &ci, &tech) == FALSE)
 		return;
 
+	if (status != 1 && status != 5) {
+		g_at_chat_register(control, "+CGREG:",
+					cgreg_notify, FALSE, NULL, NULL);
+		return;
+	}
+
 	setup_context(status);
 }
 
@@ -370,7 +356,18 @@ static void attached_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		return;
 
 	g_at_chat_send(control, "AT+CGREG?", cgreg_prefix,
-			cgreg_cb, NULL, NULL);
+						cgreg_cb, NULL, NULL);
+}
+
+static void activate_gprs(int status)
+{
+	state = STATE_ATTACHING;
+	g_print("Registered to network, roaming=%s\n",
+					status == 5 ? "true" : "false");
+
+	g_print("Activating GPRS network...\n");
+	g_at_chat_send(control, "AT+CGATT=1", none_prefix,
+						attached_cb, NULL, NULL);
 }
 
 static void creg_notify(GAtResult *result, gpointer user_data)
@@ -381,17 +378,33 @@ static void creg_notify(GAtResult *result, gpointer user_data)
 		return;
 
 	if (at_util_parse_reg_unsolicited(result, "+CREG:", &status,
-				&lac, &ci, &tech) == FALSE)
+						&lac, &ci, &tech) == FALSE)
 		return;
 
-	if (status == 1 || status == 5) {
-		g_print("Registered to network, roaming=%s\n",
-				status == 5 ? "True" : "False");
+	if (status != 1 && status != 5)
+		return;
 
-		g_print("Activating gprs network...\n");
-		g_at_chat_send(control, "AT+CGATT=1", none_prefix,
-				attached_cb, NULL, NULL);
+	activate_gprs(status);
+}
+
+static void creg_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	int status, lac, ci, tech;
+
+	if (!ok)
+		return;
+
+	if (at_util_parse_reg(result, "+CREG:", NULL, &status,
+						&lac, &ci, &tech) == FALSE)
+		return;
+
+	if (status != 1 && status != 5) {
+		g_at_chat_register(control, "+CREG:",
+						creg_notify, FALSE, NULL, NULL);
+		return;
 	}
+
+	activate_gprs(status);
 }
 
 static void register_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -403,25 +416,56 @@ static void register_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 	state = STATE_REGISTERING;
 	g_print("Waiting for network registration...\n");
+
+	g_at_chat_send(control, "AT+CREG?", creg_prefix,
+						creg_cb, NULL, NULL);
 }
 
 static void start_dial(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	if (!ok) {
+		g_print("Checking PIN status failed\n");
+		exit(1);
+	}
+
+	g_at_chat_send(control, "AT+CREG=2", none_prefix, NULL, NULL, NULL);
+	g_at_chat_send(control, "AT+CGREG=2", none_prefix, NULL, NULL, NULL);
+
+	g_at_chat_send(control, "AT+COPS=0", none_prefix,
+						register_cb, NULL, NULL);
+}
+
+static void check_pin(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	if (!ok) {
 		g_print("Turning on the modem failed\n");
 		exit(1);
 	}
 
-	g_at_chat_register(control, "+CREG:",
-				creg_notify, FALSE, NULL, NULL);
-	g_at_chat_register(control, "+CGREG:",
-				cgreg_notify, FALSE, NULL, NULL);
+	g_at_chat_send(control, "AT+CPIN?", NULL, start_dial, NULL, NULL);
+}
 
-	g_at_chat_send(control, "AT+CREG=2", none_prefix, NULL, NULL, NULL);
-	g_at_chat_send(control, "AT+CGREG=2", none_prefix, NULL, NULL, NULL);
+static void check_mode(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	GAtResultIter iter;
 
-	g_at_chat_send(control, "AT+COPS=0", none_prefix,
-			register_cb, NULL, NULL);
+	if (!ok) {
+		g_print("Checking modem mode failed\n");
+		exit(1);
+	}
+
+	g_at_result_iter_init(&iter, result);
+	g_at_result_iter_next(&iter, "+CFUN:");
+	g_at_result_iter_next_number(&iter, &oldmode);
+
+	g_print("Current modem mode is %d\n", oldmode);
+
+	if (oldmode == 1) {
+		check_pin(ok, result, user_data);
+		return;
+	}
+
+	g_at_chat_send(control, "AT+CFUN=1", NULL, check_pin, NULL, NULL);
 }
 
 static int open_serial()
@@ -527,12 +571,12 @@ static GOptionEntry options[] = {
 				"Specify CFUN offmode" },
 	{ "legacy", 'l', 0, G_OPTION_ARG_NONE, &option_legacy,
 				"Use ATD*99***<cid>#" },
-	{ "ppp", 'P', 0, G_OPTION_ARG_NONE, &option_ppp,
-				"Connect using PPP" },
 	{ "username", 'u', 0, G_OPTION_ARG_STRING, &option_username,
 				"Specify PPP username" },
 	{ "password", 'w', 0, G_OPTION_ARG_STRING, &option_password,
-				"Specifiy PPP password" },
+				"Specify PPP password" },
+	{ "pppdump", 'D', 0, G_OPTION_ARG_STRING, &option_pppdump,
+				"Specify pppdump filename" },
 	{ NULL },
 };
 
@@ -616,15 +660,20 @@ int main(int argc, char **argv)
 	event_loop = g_main_loop_new(NULL, FALSE);
 
 	g_at_chat_send(control, "ATE0Q0V1", NULL, NULL, NULL, NULL);
-	g_at_chat_send(control, "AT+CFUN=1", NULL, start_dial, NULL, NULL);
+	g_at_chat_send(control, "AT+CFUN?", cfun_prefix,
+						check_mode, NULL, NULL);
 
 	g_main_loop_run(event_loop);
 	g_source_remove(signal_source);
 	g_main_loop_unref(event_loop);
 
 out:
-	g_at_chat_unref(control);
-	g_at_chat_unref(modem);
+	if (ppp == NULL) {
+		g_at_chat_unref(control);
+		g_at_chat_unref(modem);
+	} else
+		g_at_ppp_unref(ppp);
+
 	g_free(option_apn);
 
 	return 0;
