@@ -2,7 +2,7 @@
  *  oFono - Open Source Telephony
  *
  *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
- *  Copyright (C) 2010  Gustavo F. Padovan <padovan@profusion.mobi>
+ *  Copyright (C) 2010  ProFUSION embedded systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -52,6 +52,7 @@
 #define	BLUEZ_GATEWAY_INTERFACE		BLUEZ_SERVICE ".HandsfreeGateway"
 
 #define HFP_AGENT_INTERFACE "org.bluez.HandsfreeAgent"
+#define HFP_AGENT_ERROR_INTERFACE "org.bluez.Error"
 
 #define HFP_AG_UUID	"0000111F-0000-1000-8000-00805F9B34FB"
 
@@ -65,15 +66,28 @@ static const char *cmer_prefix[] = { "+CMER:", NULL };
 static const char *chld_prefix[] = { "+CHLD:", NULL };
 
 static DBusConnection *connection;
-
-static int hfp_disable(struct ofono_modem *modem);
-static void hfp_remove(struct ofono_modem *modem);
+static GHashTable *uuid_hash = NULL;
+static GHashTable *adapter_address_hash;
 
 static void hfp_debug(const char *str, void *user_data)
 {
 	ofono_info("%s", str);
 }
 
+static void clear_data(struct ofono_modem *modem)
+{
+	struct hfp_data *data = ofono_modem_get_data(modem);
+
+	if (!data->chat)
+		return;
+
+	g_at_chat_unref(data->chat);
+	data->chat = NULL;
+
+	memset(data->cind_val, 0, sizeof(data->cind_val));
+	memset(data->cind_pos, 0, sizeof(data->cind_pos));
+}
+
 static void sevice_level_conn_established(struct ofono_modem *modem)
 {
 	DBusMessage *msg;
@@ -84,12 +98,30 @@ static void sevice_level_conn_established(struct ofono_modem *modem)
 	msg = dbus_message_new_method_return(data->slc_msg);
 	g_dbus_send_message(connection, msg);
 	dbus_message_unref(data->slc_msg);
+	data->slc_msg = NULL;
 
 	ofono_info("Service level connection established");
 
 	g_at_chat_send(data->chat, "AT+CMEE=1", NULL, NULL, NULL, NULL);
 }
 
+static void service_level_conn_failed(struct ofono_modem *modem)
+{
+	struct hfp_data *data = ofono_modem_get_data(modem);
+	DBusMessage *msg;
+
+	msg = g_dbus_create_error(data->slc_msg, HFP_AGENT_ERROR_INTERFACE
+					".Failed",
+					"HFP Handshake failed");
+	g_dbus_send_message(connection, msg);
+	dbus_message_unref(data->slc_msg);
+	data->slc_msg = NULL;
+
+	ofono_error("Service level connection failed");
+	ofono_modem_set_powered(modem, FALSE);
+	clear_data(modem);
+}
+
 static void chld_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
@@ -140,7 +172,7 @@ static void cmer_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct hfp_data *data = ofono_modem_get_data(modem);
 
 	if (!ok) {
-		hfp_disable(modem);
+		service_level_conn_failed(modem);
 		return;
 	}
 
@@ -153,11 +185,9 @@ static void cmer_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 static int send_method_call(const char *dest, const char *path,
 				const char *interface, const char *method,
-				DBusPendingCallNotifyFunction cb,
-				void *user_data, int type, ...)
+				int type, ...)
 {
 	DBusMessage *msg;
-	DBusPendingCall *call;
 	va_list args;
 
 	msg = dbus_message_new_method_call(dest, path, interface, method);
@@ -176,33 +206,156 @@ static int send_method_call(const char *dest, const char *path,
 
 	va_end(args);
 
-	if (!cb) {
-		g_dbus_send_message(connection, msg);
-		return 0;
+	g_dbus_send_message(connection, msg);
+	return 0;
+}
+
+static int send_method_call_with_reply(const char *dest, const char *path,
+				const char *interface, const char *method,
+				DBusPendingCallNotifyFunction cb,
+				void *user_data, DBusFreeFunction free_func,
+				int timeout, int type, ...)
+{
+	DBusMessage *msg;
+	DBusPendingCall *call;
+	va_list args;
+	int err;
+
+	msg = dbus_message_new_method_call(dest, path, interface, method);
+	if (!msg) {
+		ofono_error("Unable to allocate new D-Bus %s message", method);
+		err = -ENOMEM;
+		goto fail;
 	}
 
-	if (!dbus_connection_send_with_reply(connection, msg, &call, -1)) {
+	va_start(args, type);
+
+	if (!dbus_message_append_args_valist(msg, type, args)) {
+		va_end(args);
+		err = -EIO;
+		goto fail;
+	}
+
+	va_end(args);
+
+	if (timeout > 0)
+		timeout *=1000;
+
+	if (!dbus_connection_send_with_reply(connection, msg, &call, timeout)) {
 		ofono_error("Sending %s failed", method);
-		dbus_message_unref(msg);
-		return -EIO;
+		err = -EIO;
+		goto fail;
 	}
 
-	dbus_pending_call_set_notify(call, cb, user_data, NULL);
+	dbus_pending_call_set_notify(call, cb, user_data, free_func);
 	dbus_pending_call_unref(call);
 	dbus_message_unref(msg);
 
 	return 0;
+
+fail:
+	if (free_func && user_data)
+		free_func(user_data);
+
+	if (msg)
+		dbus_message_unref(msg);
+
+	return err;
 }
 
-static gboolean hfp_enable_timeout(gpointer user)
+typedef void (*PropertyHandler)(DBusMessageIter *iter, gpointer user_data);
+
+struct property_handler {
+	const char *property;
+	PropertyHandler callback;
+	gpointer user_data;
+};
+
+static gint property_handler_compare(gconstpointer a, gconstpointer b)
 {
-	struct ofono_modem *modem = user;
+	const struct property_handler *handler = a;
+	const char *property = b;
 
-	if (ofono_modem_get_powered(modem))
-		return FALSE;
+	return strcmp(handler->property, property);
+}
 
-	hfp_disable(modem);
-	return FALSE;
+static void parse_properties_reply(DBusMessage *reply,
+					const char *property, ...)
+{
+	va_list args;
+	GSList *prop_handlers = NULL;
+	DBusMessageIter array, dict;
+
+	va_start(args, property);
+
+	while (property != NULL) {
+		struct property_handler *handler =
+					g_new0(struct property_handler, 1);
+
+		handler->property = property;
+		handler->callback = va_arg(args, PropertyHandler);
+		handler->user_data = va_arg(args, gpointer);
+
+		property = va_arg(args, const char *);
+
+		prop_handlers = g_slist_prepend(prop_handlers, handler);
+	}
+
+	va_end(args);
+
+	if (dbus_message_iter_init(reply, &array) == FALSE)
+		goto done;
+
+	if (dbus_message_iter_get_arg_type(&array) != DBUS_TYPE_ARRAY)
+		goto done;
+
+	dbus_message_iter_recurse(&array, &dict);
+
+	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
+		DBusMessageIter entry, value;
+		const char *key;
+		GSList *l;
+
+		dbus_message_iter_recurse(&dict, &entry);
+
+		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_STRING)
+			goto done;
+
+		dbus_message_iter_get_basic(&entry, &key);
+
+		dbus_message_iter_next(&entry);
+
+		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_VARIANT)
+			goto done;
+
+		dbus_message_iter_recurse(&entry, &value);
+
+		l = g_slist_find_custom(prop_handlers, key,
+					property_handler_compare);
+
+		if (l) {
+			struct property_handler *handler = l->data;
+
+			handler->callback(&value, handler->user_data);
+		}
+
+		dbus_message_iter_next(&dict);
+	}
+
+done:
+	g_slist_foreach(prop_handlers, (GFunc)g_free, NULL);
+	g_slist_free(prop_handlers);
+}
+
+static void parse_string(DBusMessageIter *iter, gpointer user_data)
+{
+	char **str = user_data;
+	int arg_type = dbus_message_iter_get_arg_type(iter);
+
+	if (arg_type != DBUS_TYPE_OBJECT_PATH && arg_type != DBUS_TYPE_STRING)
+		return;
+
+	dbus_message_iter_get_basic(iter, str);
 }
 
 static void cind_status_cb(gboolean ok, GAtResult *result,
@@ -242,7 +395,7 @@ static void cind_status_cb(gboolean ok, GAtResult *result,
 	return;
 
 error:
-	hfp_disable(modem);
+	service_level_conn_failed(modem);
 }
 
 static void cind_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -302,7 +455,7 @@ static void cind_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	return;
 
 error:
-	hfp_disable(modem);
+	service_level_conn_failed(modem);
 }
 
 static void brsf_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -326,7 +479,15 @@ static void brsf_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	return;
 
 error:
-	hfp_disable(modem);
+	service_level_conn_failed(modem);
+}
+
+static void hfp_disconnected_cb(gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+
+	ofono_modem_set_powered(modem, FALSE);
+	clear_data(modem);
 }
 
 /* either oFono or Phone could request SLC connection */
@@ -353,10 +514,12 @@ static int service_level_connection(struct ofono_modem *modem, int fd)
 	if (!chat)
 		return -ENOMEM;
 
+	g_at_chat_set_disconnect_function(chat, hfp_disconnected_cb, modem);
+
 	if (getenv("OFONO_AT_DEBUG"))
 		g_at_chat_set_debug(chat, hfp_debug, NULL);
 
-	sprintf(buf, "AT+BRSF=%d", data->hf_features);
+	snprintf(buf, sizeof(buf), "AT+BRSF=%d", data->hf_features);
 	g_at_chat_send(chat, buf, brsf_prefix,
 				brsf_cb, modem, NULL);
 	data->chat = chat;
@@ -389,6 +552,11 @@ static DBusMessage *hfp_agent_release(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct ofono_modem *modem = data;
+	struct hfp_data *hfp_data = ofono_modem_get_data(modem);
+	const char *obj_path = ofono_modem_get_path(modem);
+
+	g_dbus_unregister_interface(connection, obj_path, HFP_AGENT_INTERFACE);
+	hfp_data->agent_registered = FALSE;
 
 	ofono_modem_remove(modem);
 
@@ -402,18 +570,49 @@ static GDBusMethodTable agent_methods[] = {
 	{ NULL, NULL, NULL, NULL }
 };
 
-static int hfp_create_modem(const char *device)
+static void create_path(const char *dev_addr, const char *adapter_addr,
+			char *buf, int size)
+{
+	int i, j;
+
+	for (i = 0, j = 0; adapter_addr[j] && i < size - 1; j++)
+		if (adapter_addr[j] >= '0' && adapter_addr[j] <= '9')
+			buf[i++] = adapter_addr[j];
+		else if (adapter_addr[j] >= 'A' && adapter_addr[j] <= 'F')
+			buf[i++] = adapter_addr[j];
+
+	if (i < size - 1)
+		buf[i++] = '_';
+
+	for (j = 0; dev_addr[j] && i < size - 1; j++)
+		if (dev_addr[j] >= '0' && dev_addr[j] <= '9')
+			buf[i++] = dev_addr[j];
+		else if (dev_addr[j] >= 'A' && dev_addr[j] <= 'F')
+			buf[i++] = dev_addr[j];
+
+	buf[i] = '\0';
+}
+
+static int hfp_create_modem(const char *device, const char *dev_addr,
+				const char *adapter_addr, const char *alias)
 {
 	struct ofono_modem *modem;
 	struct hfp_data *data;
+	char buf[256];
 
-	ofono_info("Using device: %s", device);
+	ofono_info("Using device: %s, devaddr: %s, adapter: %s",
+			device, dev_addr, adapter_addr);
 
-	modem = ofono_modem_create(NULL, "hfp");
+	strcpy(buf, "hfp/");
+	create_path(dev_addr, adapter_addr, buf + 4, sizeof(buf) - 4);
+
+	modem = ofono_modem_create(buf, "hfp");
+	if (modem == NULL)
+		return -ENOMEM;
 
 	data = g_try_new0(struct hfp_data, 1);
 	if (!data)
-		return -ENOMEM;
+		goto free;
 
 	data->hf_features |= HF_FEATURE_3WAY;
 	data->hf_features |= HF_FEATURE_CLIP;
@@ -422,96 +621,62 @@ static int hfp_create_modem(const char *device)
 	data->hf_features |= HF_FEATURE_ENHANCED_CALL_CONTROL;
 
 	data->handsfree_path = g_strdup(device);
+	if (data->handsfree_path == NULL)
+		goto free;
 
 	ofono_modem_set_data(modem, data);
-
+	ofono_modem_set_name(modem, alias);
 	ofono_modem_register(modem);
 
+	g_hash_table_insert(uuid_hash, g_strdup(device), modem);
+
 	return 0;
+
+free:
+	g_free(data);
+	ofono_modem_remove(modem);
+
+	return -ENOMEM;
 }
 
-static void parse_uuids(DBusMessageIter *i, const char *device)
+static void has_hfp_uuid(DBusMessageIter *array, gpointer user_data)
 {
-	DBusMessageIter variant, ai;
-	const char *value;
+	gboolean *hfp = user_data;
+	DBusMessageIter value;
 
-	dbus_message_iter_recurse(i, &variant);
-	dbus_message_iter_recurse(&variant, &ai);
+	if (dbus_message_iter_get_arg_type(array) != DBUS_TYPE_ARRAY)
+		return;
 
-	while (dbus_message_iter_get_arg_type(&ai) != DBUS_TYPE_INVALID) {
-		dbus_message_iter_get_basic(&ai, &value);
+	dbus_message_iter_recurse(array, &value);
 
-		if (!strcasecmp(value, HFP_AG_UUID))
-			hfp_create_modem(device);
+	while (dbus_message_iter_get_arg_type(&value) == DBUS_TYPE_STRING) {
+		const char *uuid;
 
-		if (!dbus_message_iter_next(&ai))
+		dbus_message_iter_get_basic(&value, &uuid);
+
+		if (!strcasecmp(uuid, HFP_AG_UUID)) {
+			*hfp = TRUE;
 			return;
-	}
-}
-
-static void parse_get_properties(DBusMessage *reply, const char *device)
-{
-	DBusMessageIter arg, element, variant;
-	const char *key;
-
-	if (!dbus_message_iter_init(reply, &arg)) {
-		ofono_debug("GetProperties reply has no arguments.");
-		return;
-	}
-
-	if (dbus_message_iter_get_arg_type(&arg) != DBUS_TYPE_ARRAY) {
-		ofono_debug("GetProperties argument is not an array.");
-		return;
-	}
-
-	dbus_message_iter_recurse(&arg, &element);
-
-	while (dbus_message_iter_get_arg_type(&element) != DBUS_TYPE_INVALID) {
-		if (dbus_message_iter_get_arg_type(&element) ==
-				DBUS_TYPE_DICT_ENTRY) {
-			DBusMessageIter dict;
-
-			dbus_message_iter_recurse(&element, &dict);
-
-			if (dbus_message_iter_get_arg_type(&dict) !=
-					DBUS_TYPE_STRING) {
-				ofono_debug("Property name not a string.");
-				return;
-			}
-
-			dbus_message_iter_get_basic(&dict, &key);
-
-			if (!dbus_message_iter_next(&dict))  {
-				ofono_debug("Property value missing");
-				return;
-			}
-
-			if (dbus_message_iter_get_arg_type(&dict) !=
-					DBUS_TYPE_VARIANT) {
-				ofono_debug("Property value not a variant.");
-				return;
-			}
-
-			if (!strcmp(key, "UUIDs"))
-				parse_uuids(&dict, device);
-
-			dbus_message_iter_recurse(&dict, &variant);
 		}
 
-		if (!dbus_message_iter_next(&element))
-			return;
+		dbus_message_iter_next(&value);
 	}
 }
 
-static void get_properties_cb(DBusPendingCall *call, gpointer user_data)
+static void device_properties_cb(DBusPendingCall *call, gpointer user_data)
 {
 	DBusMessage *reply;
-	const char *device = user_data;
+	char *path = user_data;
+	gboolean have_hfp = FALSE;
+	const char *adapter = NULL;
+	const char *adapter_addr = NULL;
+	const char *device_addr = NULL;
+	const char *alias = NULL;
 
 	reply = dbus_pending_call_steal_reply(call);
 
 	if (dbus_message_is_error(reply, DBUS_ERROR_SERVICE_UNKNOWN)) {
-		ofono_debug("Bluetooth daemon is apparently not available.");
+		DBG("Bluetooth daemon is apparently not available.");
 		goto done;
 	}
 
@@ -523,54 +688,79 @@ static void get_properties_cb(DBusPendingCall *call, gpointer user_data)
 		goto done;
 	}
 
-	parse_get_properties(reply, device);
+	parse_properties_reply(reply, "UUIDs", has_hfp_uuid, &have_hfp,
+				"Adapter", parse_string, &adapter,
+				"Address", parse_string, &device_addr,
+				"Alias", parse_string, &alias, NULL);
+
+	if (adapter)
+		adapter_addr = g_hash_table_lookup(adapter_address_hash,
+							adapter);
+
+	if (have_hfp && device_addr && adapter_addr)
+		hfp_create_modem(path, device_addr, adapter_addr, alias);
 
 done:
 	dbus_message_unref(reply);
 }
 
-static void list_devices_cb(DBusPendingCall *call, gpointer user_data)
+static void parse_devices(DBusMessageIter *array, gpointer user_data)
 {
-	DBusError err;
+	DBusMessageIter value;
+	GSList **device_list = user_data;
+
+	DBG("");
+
+	if (dbus_message_iter_get_arg_type(array) != DBUS_TYPE_ARRAY)
+		return;
+
+	dbus_message_iter_recurse(array, &value);
+
+	while (dbus_message_iter_get_arg_type(&value)
+			== DBUS_TYPE_OBJECT_PATH) {
+		const char *path;
+
+		dbus_message_iter_get_basic(&value, &path);
+
+		*device_list = g_slist_prepend(*device_list, (gpointer) path);
+
+		dbus_message_iter_next(&value);
+	}
+}
+
+static void adapter_properties_cb(DBusPendingCall *call, gpointer user_data)
+{
+	const char *path = user_data;
 	DBusMessage *reply;
-	const char **device = NULL;
-	int num, ret, i;
+	GSList *device_list = NULL;
+	GSList *l;
+	const char *addr;
 
 	reply = dbus_pending_call_steal_reply(call);
 
 	if (dbus_message_is_error(reply, DBUS_ERROR_SERVICE_UNKNOWN)) {
-		ofono_debug("Bluetooth daemon is apparently not available.");
+		DBG("Bluetooth daemon is apparently not available.");
 		goto done;
 	}
 
-	dbus_error_init(&err);
+	parse_properties_reply(reply, "Devices", parse_devices, &device_list,
+				"Address", parse_string, &addr, NULL);
 
-	if (dbus_message_get_args(reply, &err, DBUS_TYPE_ARRAY,
-				DBUS_TYPE_OBJECT_PATH, &device,
-				&num, DBUS_TYPE_INVALID) == FALSE) {
-		if (device == NULL) {
-			dbus_error_free(&err);
-			goto done;
-		}
+	DBG("Adapter Address: %s, Path: %s", addr, path);
+	g_hash_table_insert(adapter_address_hash,
+				g_strdup(path), g_strdup(addr));
 
-		if (dbus_error_is_set(&err) == TRUE) {
-			ofono_error("%s", err.message);
-			dbus_error_free(&err);
-		}
+	for (l = device_list; l; l = l->next) {
+		const char *device = l->data;
 
-		goto done;
-	}
-
-	for (i = 0 ; i < num ; i++) {
-		ret = send_method_call(BLUEZ_SERVICE, device[i],
+		send_method_call_with_reply(BLUEZ_SERVICE, device,
 				BLUEZ_DEVICE_INTERFACE, "GetProperties",
-				get_properties_cb, (void *)device[i],
-				DBUS_TYPE_INVALID);
-		if (ret < 0)
-			ofono_error("GetProperties failed(%d)", ret);
+				device_properties_cb, g_strdup(device), g_free,
+				-1, DBUS_TYPE_INVALID);
 	}
 
 done:
+	g_slist_free(device_list);
 	dbus_message_unref(reply);
 }
 
@@ -583,71 +773,152 @@ static gboolean adapter_added(DBusConnection *connection, DBusMessage *message,
 	dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
 				DBUS_TYPE_INVALID);
 
-	ret = send_method_call(BLUEZ_SERVICE, path,
-			BLUEZ_ADAPTER_INTERFACE, "ListDevices",
-			list_devices_cb, NULL,
-			DBUS_TYPE_INVALID);
-
-	if (ret < 0)
-		ofono_error("ListDevices failed(%d)", ret);
+	ret = send_method_call_with_reply(BLUEZ_SERVICE, path,
+			BLUEZ_ADAPTER_INTERFACE, "GetProperties",
+			adapter_properties_cb, g_strdup(path), g_free,
+			-1, DBUS_TYPE_INVALID);
 
 	return TRUE;
 }
 
-static void list_adapters_cb(DBusPendingCall *call, gpointer user_data)
+static gboolean adapter_removed(DBusConnection *connection,
+				DBusMessage *message, void *user_data)
+{
+	const char *path;
+
+	if (dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+				DBUS_TYPE_INVALID) == TRUE)
+		g_hash_table_remove(adapter_address_hash, path);
+
+	return TRUE;
+}
+
+static gboolean property_changed(DBusConnection *connection, DBusMessage *msg,
+				void *user_data)
+{
+	const char *property;
+	DBusMessageIter iter;
+
+	dbus_message_iter_init(msg, &iter);
+
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
+		return FALSE;
+
+	dbus_message_iter_get_basic(&iter, &property);
+	if (g_str_equal(property, "UUIDs") == TRUE) {
+		gboolean have_hfp = FALSE;
+		const char *path = dbus_message_get_path(msg);
+		DBusMessageIter variant;
+
+		/* We already have this device in our hash, ignore */
+		if (g_hash_table_lookup(uuid_hash, path) != NULL)
+			return TRUE;
+
+		if (!dbus_message_iter_next(&iter))
+			return FALSE;
+
+		if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
+			return FALSE;
+
+		dbus_message_iter_recurse(&iter, &variant);
+
+		has_hfp_uuid(&variant, &have_hfp);
+
+		/* We need the full set of properties to be able to create
+		 * the modem properly, including Adapter and Alias, so
+		 * refetch everything again
+		 */
+		if (have_hfp)
+			send_method_call_with_reply(BLUEZ_SERVICE, path,
+				BLUEZ_DEVICE_INTERFACE, "GetProperties",
+				device_properties_cb, g_strdup(path), g_free,
+				-1, DBUS_TYPE_INVALID);
+	} else if (g_str_equal(property, "Alias") == TRUE) {
+		const char *path = dbus_message_get_path(msg);
+		struct ofono_modem *modem =
+			g_hash_table_lookup(uuid_hash, path);
+		const char *alias = NULL;
+		DBusMessageIter variant;
+
+		if (modem == NULL)
+			return TRUE;
+
+		if (!dbus_message_iter_next(&iter))
+			return FALSE;
+
+		if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
+			return FALSE;
+
+		dbus_message_iter_recurse(&iter, &variant);
+
+		parse_string(&variant, &alias);
+
+		ofono_modem_set_name(modem, alias);
+	}
+
+	return TRUE;
+}
+
+static void parse_adapters(DBusMessageIter *array, gpointer user_data)
+{
+	DBusMessageIter value;
+
+	DBG("");
+
+	if (dbus_message_iter_get_arg_type(array) != DBUS_TYPE_ARRAY)
+		return;
+
+	dbus_message_iter_recurse(array, &value);
+
+	while (dbus_message_iter_get_arg_type(&value)
+			== DBUS_TYPE_OBJECT_PATH) {
+		const char *path;
+
+		dbus_message_iter_get_basic(&value, &path);
+
+		DBG("Calling GetProperties on %s", path);
+
+		send_method_call_with_reply(BLUEZ_SERVICE, path,
+				BLUEZ_ADAPTER_INTERFACE, "GetProperties",
+				adapter_properties_cb, g_strdup(path), g_free,
+				-1, DBUS_TYPE_INVALID);
+
+		dbus_message_iter_next(&value);
+	}
+}
+
+static void manager_properties_cb(DBusPendingCall *call, gpointer user_data)
 {
-	DBusError err;
 	DBusMessage *reply;
-	char **adapter = NULL;
-	int num, ret, i;
 
 	reply = dbus_pending_call_steal_reply(call);
 
 	if (dbus_message_is_error(reply, DBUS_ERROR_SERVICE_UNKNOWN)) {
-		ofono_debug("Bluetooth daemon is apparently not available.");
+		DBG("Bluetooth daemon is apparently not available.");
 		goto done;
 	}
 
-	dbus_error_init(&err);
-
-	if (dbus_message_get_args(reply, &err, DBUS_TYPE_ARRAY,
-				DBUS_TYPE_OBJECT_PATH, &adapter,
-				&num, DBUS_TYPE_INVALID) == FALSE) {
-		if (adapter == NULL) {
-			dbus_error_free(&err);
-			goto done;
-		}
-
-		if (dbus_error_is_set(&err) == TRUE) {
-			ofono_error("%s", err.message);
-			dbus_error_free(&err);
-		}
-
-		goto done;
-	}
-
-	for (i = 0 ; i < num ; i++) {
-		ret = send_method_call(BLUEZ_SERVICE, adapter[i],
-				BLUEZ_ADAPTER_INTERFACE, "ListDevices",
-				list_devices_cb, NULL,
-				DBUS_TYPE_INVALID);
-
-		if (ret < 0)
-			ofono_error("ListDevices failed(%d)", ret);
-	}
-
-	g_strfreev(adapter);
+	parse_properties_reply(reply, "Adapters", parse_adapters, NULL, NULL);
 
 done:
 	dbus_message_unref(reply);
 }
 
-static int hfp_load_modems()
+static gboolean hfp_remove_each_modem(gpointer key, gpointer value, gpointer user_data)
 {
-	return send_method_call(BLUEZ_SERVICE, "/",
-				BLUEZ_MANAGER_INTERFACE, "ListAdapters",
-				list_adapters_cb, NULL,
-				DBUS_TYPE_INVALID);
+	struct ofono_modem *modem = value;
+
+	ofono_modem_remove(modem);
+
+	return TRUE;
+}
+
+static void bluetooth_disconnect(DBusConnection *connection, void *user_data)
+{
+	if (uuid_hash == NULL)
+		return;
+
+	g_hash_table_foreach_remove(uuid_hash, hfp_remove_each_modem, NULL);
 }
 
 static int hfp_register_ofono_handsfree(struct ofono_modem *modem)
@@ -655,15 +926,12 @@ static int hfp_register_ofono_handsfree(struct ofono_modem *modem)
 	const char *obj_path = ofono_modem_get_path(modem);
 	struct hfp_data *data = ofono_modem_get_data(modem);
 
-	ofono_debug("Registering oFono Agent to bluetooth daemon");
-
-	if (!data->handsfree_path)
-		return -EINVAL;
+	DBG("Registering oFono Agent to bluetooth daemon");
 
 	return send_method_call(BLUEZ_SERVICE, data->handsfree_path,
 				BLUEZ_GATEWAY_INTERFACE, "RegisterAgent",
-				NULL, NULL, DBUS_TYPE_OBJECT_PATH,
-				&obj_path, DBUS_TYPE_INVALID);
+				DBUS_TYPE_OBJECT_PATH, &obj_path,
+				DBUS_TYPE_INVALID);
 }
 
 static int hfp_unregister_ofono_handsfree(struct ofono_modem *modem)
@@ -671,24 +939,27 @@ static int hfp_unregister_ofono_handsfree(struct ofono_modem *modem)
 	const char *obj_path = ofono_modem_get_path(modem);
 	struct hfp_data *data = ofono_modem_get_data(modem);
 
-	ofono_debug("Unregistering oFono Agent from bluetooth daemon");
-
-	if (!data->handsfree_path)
-		return -EINVAL;
+	DBG("Unregistering oFono Agent from bluetooth daemon");
 
 	return send_method_call(BLUEZ_SERVICE, data->handsfree_path,
 				BLUEZ_GATEWAY_INTERFACE, "UnregisterAgent",
-				NULL, NULL, DBUS_TYPE_OBJECT_PATH,
-				&obj_path, DBUS_TYPE_INVALID);
+				DBUS_TYPE_OBJECT_PATH, &obj_path,
+				DBUS_TYPE_INVALID);
 }
 
 static int hfp_probe(struct ofono_modem *modem)
 {
 	const char *obj_path = ofono_modem_get_path(modem);
+	struct hfp_data *data = ofono_modem_get_data(modem);
+
+	if (!data)
+		return -EINVAL;
 
 	g_dbus_register_interface(connection, obj_path, HFP_AGENT_INTERFACE,
 			agent_methods, NULL, NULL, modem, NULL);
 
+	data->agent_registered = TRUE;
+
 	if (hfp_register_ofono_handsfree(modem) != 0)
 		return -EINVAL;
 
@@ -698,81 +969,117 @@ static int hfp_probe(struct ofono_modem *modem)
 static void hfp_remove(struct ofono_modem *modem)
 {
 	struct hfp_data *data = ofono_modem_get_data(modem);
+	const char *obj_path = ofono_modem_get_path(modem);
 
-	hfp_unregister_ofono_handsfree(modem);
+	if (g_dbus_unregister_interface(connection, obj_path,
+					HFP_AGENT_INTERFACE))
+		hfp_unregister_ofono_handsfree(modem);
 
-	if (data->handsfree_path)
-		g_free(data->handsfree_path);
+	g_hash_table_remove(uuid_hash, data->handsfree_path);
 
-	if (data)
-		g_free(data);
+	g_free(data->handsfree_path);
+	g_free(data);
 
 	ofono_modem_set_data(modem, NULL);
 }
 
-static int hfp_connect_ofono_handsfree(struct ofono_modem *modem)
+static void hfp_connect_reply(DBusPendingCall *call, gpointer user_data)
 {
+	struct ofono_modem *modem = user_data;
 	struct hfp_data *data = ofono_modem_get_data(modem);
+	DBusError derr;
+	DBusMessage *reply;
+	int ret;
 
-	ofono_debug("Connect to bluetooth daemon");
+	reply = dbus_pending_call_steal_reply(call);
 
-	if (!data->handsfree_path || !connection)
-		return -EINVAL;
+	if (ofono_modem_get_powered(modem))
+		goto done;
 
-	return send_method_call(BLUEZ_SERVICE, data->handsfree_path,
-				BLUEZ_GATEWAY_INTERFACE, "Connect",
-				NULL, NULL, DBUS_TYPE_INVALID);
+	dbus_error_init(&derr);
+	if (!dbus_set_error_from_message(&derr, reply))
+		goto done;
+
+	DBG("Connect reply: %s", derr.message);
+
+	if (dbus_error_has_name(&derr, DBUS_ERROR_NO_REPLY)) {
+		ret = send_method_call(BLUEZ_SERVICE, data->handsfree_path,
+					BLUEZ_GATEWAY_INTERFACE, "Disconnect",
+					DBUS_TYPE_INVALID);
+		if (ret < 0)
+			ofono_error("Disconnect failed(%d)", ret);
+	}
+
+	ofono_modem_set_powered(modem, FALSE);
+
+	dbus_error_free(&derr);
+
+done:
+	dbus_message_unref(reply);
 }
 
 /* power up hardware */
 static int hfp_enable(struct ofono_modem *modem)
 {
 	struct hfp_data *data = ofono_modem_get_data(modem);
+	int status;
 
 	DBG("%p", modem);
 
-	data->at_timeout =
-		g_timeout_add_seconds(10, hfp_enable_timeout, modem);
+	status = send_method_call_with_reply(BLUEZ_SERVICE,
+				data->handsfree_path,
+				BLUEZ_GATEWAY_INTERFACE, "Connect",
+				hfp_connect_reply, modem, NULL,
+				15, DBUS_TYPE_INVALID);
 
-	if (hfp_connect_ofono_handsfree(modem) < 0)
+	if (status < 0)
 		return -EINVAL;
 
 	return -EINPROGRESS;
 }
 
-static int hfp_disconnect_ofono_handsfree(struct ofono_modem *modem)
+static void hfp_power_down(DBusPendingCall *call, gpointer user_data)
 {
-	struct hfp_data *data = ofono_modem_get_data(modem);
+	struct ofono_modem *modem = user_data;
+	DBusMessage *reply;
+	DBusError derr;
 
-	if (!data->handsfree_path || !connection)
-		return -EINVAL;
+	reply = dbus_pending_call_steal_reply(call);
 
-	return send_method_call(BLUEZ_SERVICE, data->handsfree_path,
-				BLUEZ_GATEWAY_INTERFACE, "Disconnect",
-				NULL, NULL, DBUS_TYPE_INVALID);
+	dbus_error_init(&derr);
+	if (dbus_set_error_from_message(&derr, reply)) {
+		DBG("Disconnect reply: %s", derr.message);
+		dbus_error_free(&derr);
+		goto done;
+	}
+
+	ofono_modem_set_powered(modem, FALSE);
+
+done:
+	dbus_message_unref(reply);
 }
 
 static int hfp_disable(struct ofono_modem *modem)
 {
 	struct hfp_data *data = ofono_modem_get_data(modem);
+	int status;
 
 	DBG("%p", modem);
 
-	if (!data->chat)
-		return 0;
+	clear_data(modem);
 
-	g_at_chat_unref(data->chat);
-	data->chat = NULL;
+	if (data->agent_registered) {
+		status = send_method_call_with_reply(BLUEZ_SERVICE,
+					data->handsfree_path,
+					BLUEZ_GATEWAY_INTERFACE, "Disconnect",
+					hfp_power_down, modem, NULL, 15,
+					DBUS_TYPE_INVALID);
 
-	memset(data->cind_val, 0, sizeof(data->cind_val));
-	memset(data->cind_pos, 0, sizeof(data->cind_pos));
+		if (status < 0)
+			return -EINVAL;
+	}
 
-	g_source_remove(data->at_timeout);
-
-	ofono_modem_set_powered(modem, FALSE);
-
-	hfp_disconnect_ofono_handsfree(modem);
-	return 0;
+	return -EINPROGRESS;
 }
 
 static void hfp_pre_sim(struct ofono_modem *modem)
@@ -801,9 +1108,12 @@ static struct ofono_modem_driver hfp_driver = {
 	.post_sim	= hfp_post_sim,
 };
 
-static guint added_watch;
+static guint bluetooth_exit_watch;
+static guint adapter_added_watch;
+static guint adapter_removed_watch;
+static guint uuid_watch;
 
-static int hfp_init(void)
+static int hfp_init()
 {
 	int err;
 
@@ -812,37 +1122,73 @@ static int hfp_init(void)
 
 	connection = ofono_dbus_get_connection();
 
-	added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+	bluetooth_exit_watch = g_dbus_add_service_watch(connection, BLUEZ_SERVICE,
+			NULL, bluetooth_disconnect, NULL, NULL);
+
+	adapter_added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
 						BLUEZ_MANAGER_INTERFACE,
 						"AdapterAdded",
 						adapter_added, NULL, NULL);
 
-	if (added_watch == 0) {
+	adapter_removed_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_MANAGER_INTERFACE,
+						"AdapterRemoved",
+						adapter_removed, NULL, NULL);
+
+	uuid_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_DEVICE_INTERFACE,
+						"PropertyChanged",
+						property_changed, NULL, NULL);
+
+	if (bluetooth_exit_watch == 0 || adapter_added_watch == 0 ||
+			adapter_removed_watch == 0|| uuid_watch == 0) {
 		err = -EIO;
 		goto remove;
 	}
 
+	uuid_hash = g_hash_table_new_full(g_str_hash, g_str_equal,
+						g_free, NULL);
+
+	adapter_address_hash = g_hash_table_new_full(g_str_hash, g_str_equal,
+							g_free, g_free);
+
 	err = ofono_modem_driver_register(&hfp_driver);
 	if (err < 0)
 		goto remove;
 
-	hfp_load_modems();
+	send_method_call_with_reply(BLUEZ_SERVICE, "/",
+				BLUEZ_MANAGER_INTERFACE, "GetProperties",
+				manager_properties_cb, NULL, NULL, -1,
+				DBUS_TYPE_INVALID);
 
 	return 0;
 
 remove:
-	g_dbus_remove_watch(connection, added_watch);
+	g_dbus_remove_watch(connection, bluetooth_exit_watch);
+	g_dbus_remove_watch(connection, adapter_added_watch);
+	g_dbus_remove_watch(connection, adapter_removed_watch);
+	g_dbus_remove_watch(connection, uuid_watch);
 
-	dbus_connection_unref(connection);
+	if (uuid_hash)
+		g_hash_table_destroy(uuid_hash);
+
+	if (adapter_address_hash)
+		g_hash_table_destroy(adapter_address_hash);
 
 	return err;
 }
 
-static void hfp_exit(void)
+static void hfp_exit()
 {
-	g_dbus_remove_watch(connection, added_watch);
+	g_dbus_remove_watch(connection, bluetooth_exit_watch);
+	g_dbus_remove_watch(connection, adapter_added_watch);
+	g_dbus_remove_watch(connection, adapter_removed_watch);
+	g_dbus_remove_watch(connection, uuid_watch);
 
 	ofono_modem_driver_unregister(&hfp_driver);
+
+	g_hash_table_destroy(uuid_hash);
+	g_hash_table_destroy(adapter_address_hash);
 }
 
 OFONO_PLUGIN_DEFINE(hfp, "Hands-Free Profile Plugins", VERSION,
