@@ -36,6 +36,7 @@
 #include <ofono/log.h>
 
 static GSList *modem_list = NULL;
+static GHashTable *devpath_list = NULL;
 
 static struct ofono_modem *find_modem(const char *devpath)
 {
@@ -88,6 +89,24 @@ static const char *get_serial(struct udev_device *udev_device)
 	return serial;
 }
 
+static const char *get_usb_num(struct udev_device *udev_device)
+{
+	struct udev_list_entry *entry;
+	const char *num = NULL;
+
+	entry = udev_device_get_properties_list_entry(udev_device);
+	while (entry) {
+		const char *name = udev_list_entry_get_name(entry);
+
+		if (g_strcmp0(name, "ID_USB_INTERFACE_NUM") == 0)
+			num = udev_list_entry_get_value(entry);
+
+		entry = udev_list_entry_get_next(entry);
+	}
+
+	return num;
+}
+
 #define MODEM_DEVICE		"ModemDevice"
 #define DATA_DEVICE		"DataDevice"
 #define GPS_DEVICE		"GPSDevice"
@@ -97,7 +116,7 @@ static void add_mbm(struct ofono_modem *modem,
 					struct udev_device *udev_device)
 {
 	const char *desc, *devnode;
-	const char *device, *network;
+	const char *device, *data, *network;
 	int registered;
 
 	desc = udev_device_get_sysattr_value(udev_device, "device/interface");
@@ -112,11 +131,15 @@ static void add_mbm(struct ofono_modem *modem,
 		return;
 
 	if (g_str_has_suffix(desc, "Minicard Modem") ||
+			g_str_has_suffix(desc, "Minicard Modem 2") ||
 			g_str_has_suffix(desc, "Mini-Card Modem") ||
 			g_str_has_suffix(desc, "Broadband Modem") ||
 			g_str_has_suffix(desc, "Broadband USB Modem")) {
 		devnode = udev_device_get_devnode(udev_device);
-		ofono_modem_set_string(modem, MODEM_DEVICE, devnode);
+		if (ofono_modem_get_string(modem, MODEM_DEVICE) == NULL)
+			ofono_modem_set_string(modem, MODEM_DEVICE, devnode);
+		else
+			ofono_modem_set_string(modem, DATA_DEVICE, devnode);
 	} else if (g_str_has_suffix(desc, "Minicard Data Modem") ||
 			g_str_has_suffix(desc, "Mini-Card Data Modem") ||
 			g_str_has_suffix(desc, "Broadband Data Modem")) {
@@ -139,9 +162,10 @@ static void add_mbm(struct ofono_modem *modem,
 	}
 
 	device  = ofono_modem_get_string(modem, MODEM_DEVICE);
+	data = ofono_modem_get_string(modem, DATA_DEVICE);
 	network = ofono_modem_get_string(modem, NETWORK_INTERFACE);
 
-	if (device != NULL && network != NULL) {
+	if (device != NULL && data != NULL && network != NULL) {
 		ofono_modem_set_integer(modem, "Registered", 1);
 		ofono_modem_register(modem);
 	}
@@ -195,18 +219,45 @@ static void add_hso(struct ofono_modem *modem,
 static void add_huawei(struct ofono_modem *modem,
 					struct udev_device *udev_device)
 {
-	const char *devnode;
-	int registered;
+	const char *devnode, *num;
+	int primary, secondary;
 
-	registered = ofono_modem_get_integer(modem, "Registered");
-	if (registered != 0)
+	primary = ofono_modem_get_integer(modem, "PrimaryRegistered");
+	secondary = ofono_modem_get_integer(modem, "SecondaryRegistered");
+
+	if (primary && secondary)
 		return;
 
-	devnode = udev_device_get_devnode(udev_device);
-	ofono_modem_set_string(modem, "Device", devnode);
+	num = get_usb_num(udev_device);
 
-	ofono_modem_set_integer(modem, "Registered", 1);
-	ofono_modem_register(modem);
+	/*
+	 * Here is is assumed that that usb port number 0 is the control
+	 * port and port 2 is the event port. This assumption will surely
+	 * be false with some devices and better heuristics is needed.
+	 */
+	if (g_strcmp0(num, "00") == 0) {
+		if (primary != 0)
+			return;
+
+		devnode = udev_device_get_devnode(udev_device);
+		ofono_modem_set_string(modem, "Device", devnode);
+
+		primary = 1;
+		ofono_modem_set_integer(modem, "PrimaryRegistered", primary);
+	} else if (g_strcmp0(num, "02") == 0) {
+		if (secondary != 0)
+			return;
+
+		devnode = udev_device_get_devnode(udev_device);
+		ofono_modem_set_string(modem, "SecondaryDevice", devnode);
+
+		secondary = 1;
+		ofono_modem_set_integer(modem, "SecondaryRegistered",
+					secondary);
+	}
+
+	if (primary && secondary)
+		ofono_modem_register(modem);
 }
 
 static void add_em770(struct ofono_modem *modem,
@@ -254,7 +305,7 @@ static void add_modem(struct udev_device *udev_device)
 {
 	struct ofono_modem *modem;
 	struct udev_device *parent;
-	const char *devpath, *driver;
+	const char *devpath, *curpath, *driver;
 
 	parent = udev_device_get_parent(udev_device);
 	if (parent == NULL)
@@ -290,6 +341,12 @@ static void add_modem(struct udev_device *udev_device)
 		modem_list = g_slist_prepend(modem_list, modem);
 	}
 
+	curpath = udev_device_get_devpath(udev_device);
+	if (curpath == NULL)
+		return;
+
+	g_hash_table_insert(devpath_list, g_strdup(curpath), g_strdup(devpath));
+
 	if (g_strcmp0(driver, "mbm") == 0)
 		add_mbm(modem, udev_device);
 	else if (g_strcmp0(driver, "hso") == 0)
@@ -302,30 +359,25 @@ static void add_modem(struct udev_device *udev_device)
 		add_novatel(modem, udev_device);
 }
 
+static gboolean devpath_remove(gpointer key, gpointer value, gpointer user_data)
+{
+	const char *path = value;
+	const char *devpath = user_data;
+
+	return g_str_equal(path, devpath);
+}
+
 static void remove_modem(struct udev_device *udev_device)
 {
 	struct ofono_modem *modem;
-	struct udev_device *parent;
-	const char *devpath, *driver = NULL;
+	const char *curpath = udev_device_get_devpath(udev_device);
+	char *devpath, *remove;
 
-	parent = udev_device_get_parent(udev_device);
-	if (parent == NULL)
+	if (curpath == NULL)
 		return;
 
-	driver = get_driver(parent);
-	if (driver == NULL) {
-		parent = udev_device_get_parent(parent);
-		driver = get_driver(parent);
-		if (driver == NULL) {
-			parent = udev_device_get_parent(parent);
-			driver = get_driver(parent);
-			if (driver == NULL)
-				return;
-		}
-	}
-
-	devpath = udev_device_get_devpath(parent);
-	if (devpath == NULL)
+	devpath = g_hash_table_lookup(devpath_list, curpath);
+	if (!devpath)
 		return;
 
 	modem = find_modem(devpath);
@@ -335,6 +387,12 @@ static void remove_modem(struct udev_device *udev_device)
 	modem_list = g_slist_remove(modem_list, modem);
 
 	ofono_modem_remove(modem);
+
+	remove = g_strdup(devpath);
+
+	g_hash_table_foreach_remove(devpath_list, devpath_remove, remove);
+
+	g_free(remove);
 }
 
 static void enumerate_devices(struct udev *context)
@@ -439,15 +497,24 @@ static void udev_start(void)
 
 static int udev_init(void)
 {
+	devpath_list = g_hash_table_new_full(g_str_hash, g_str_equal,
+						g_free, g_free);
+	if (!devpath_list) {
+		ofono_error("Failed to create udev path list");
+		return -ENOMEM;
+	}
+
 	udev_ctx = udev_new();
 	if (udev_ctx == NULL) {
 		ofono_error("Failed to create udev context");
+		g_hash_table_destroy(devpath_list);
 		return -EIO;
 	}
 
 	udev_mon = udev_monitor_new_from_netlink(udev_ctx, "udev");
 	if (udev_mon == NULL) {
 		ofono_error("Failed to create udev monitor");
+		g_hash_table_destroy(devpath_list);
 		udev_unref(udev_ctx);
 		udev_ctx = NULL;
 		return -EIO;
@@ -479,6 +546,9 @@ static void udev_exit(void)
 	g_slist_free(modem_list);
 	modem_list = NULL;
 
+	g_hash_table_destroy(devpath_list);
+	devpath_list = NULL;
+
 	if (udev_ctx == NULL)
 		return;
 
