@@ -39,8 +39,6 @@
 
 #define uninitialized_var(x) x = x
 
-#define SMS_MANAGER_INTERFACE "org.ofono.SmsManager"
-
 #define SMS_MANAGER_FLAG_CACHED 0x1
 
 #define SETTINGS_STORE "sms"
@@ -60,7 +58,6 @@ struct ofono_sms {
 	unsigned int next_msg_id;
 	guint ref;
 	GQueue *txq;
-	time_t last_mms;
 	gint tx_source;
 	struct ofono_message_waiting *mw;
 	unsigned int mw_watch;
@@ -105,7 +102,7 @@ static void set_sca(struct ofono_sms *sms,
 	value = phone_number_to_string(&sms->sca);
 
 	ofono_dbus_signal_property_changed(conn, path,
-						SMS_MANAGER_INTERFACE,
+						OFONO_SMS_MANAGER_INTERFACE,
 						"ServiceCenterAddress",
 						DBUS_TYPE_STRING, &value);
 }
@@ -333,7 +330,6 @@ static void tx_finished(const struct ofono_error *error, int mr, void *data)
 static gboolean tx_next(gpointer user_data)
 {
 	struct ofono_sms *sms = user_data;
-	time_t ts;
 	int send_mms = 0;
 	struct tx_queue_entry *entry = g_queue_peek_head(sms->txq);
 	struct pending_pdu *pdu = &entry->pdus[entry->cur_pdu];
@@ -348,14 +344,12 @@ static gboolean tx_next(gpointer user_data)
 	if (!entry)
 		return FALSE;
 
-	ts = time(NULL);
-
-	if ((g_queue_get_length(sms->txq) > 1) &&
-			((ts - sms->last_mms) > 60))
+	if (g_queue_get_length(sms->txq) > 1
+			|| (entry->num_pdus - entry->cur_pdu) > 1)
 		send_mms = 1;
 
-	sms->driver->submit(sms, pdu->pdu, pdu->pdu_len, pdu->tpdu_len, send_mms,
-				tx_finished, sms);
+	sms->driver->submit(sms, pdu->pdu, pdu->pdu_len, pdu->tpdu_len,
+				send_mms, tx_finished, sms);
 
 	return FALSE;
 }
@@ -400,6 +394,19 @@ static struct tx_queue_entry *create_tx_queue_entry(GSList *msg_list)
 	return entry;
 }
 
+/*
+ * Pre-process a SMS text message and deliver it [D-Bus SendMessage()]
+ *
+ * @conn: D-Bus connection
+ * @msg: message data (telephone number and text)
+ * @data: SMS object to use for transmision
+ *
+ * An alphabet is chosen for the text and it (might be) segmented in
+ * fragments by sms_text_prepare() into @msg_list. A queue list @entry
+ * is created by create_tx_queue_entry() and g_queue_push_tail()
+ * appends that entry to the SMS transmit queue. Then the tx_next()
+ * function is scheduled to run to process the queue.
+ */
 static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
@@ -506,7 +513,7 @@ static void dispatch_text_message(struct ofono_sms *sms,
 	else
 		signal_name = "IncomingMessage";
 
-	signal = dbus_message_new_signal(path, SMS_MANAGER_INTERFACE,
+	signal = dbus_message_new_signal(path, OFONO_SMS_MANAGER_INTERFACE,
 						signal_name);
 
 	if (!signal)
@@ -845,8 +852,8 @@ static void sms_unregister(struct ofono_atom *atom)
 	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
 	const char *path = __ofono_atom_get_path(atom);
 
-	g_dbus_unregister_interface(conn, path, SMS_MANAGER_INTERFACE);
-	ofono_modem_remove_interface(modem, SMS_MANAGER_INTERFACE);
+	g_dbus_unregister_interface(conn, path, OFONO_SMS_MANAGER_INTERFACE);
+	ofono_modem_remove_interface(modem, OFONO_SMS_MANAGER_INTERFACE);
 
 	if (sms->mw_watch) {
 		__ofono_modem_remove_atom_watch(modem, sms->mw_watch);
@@ -899,6 +906,17 @@ static void sms_remove(struct ofono_atom *atom)
 	g_free(sms);
 }
 
+
+/*
+ * Create a SMS driver
+ *
+ * This creates a SMS driver that is hung off a @modem
+ * object. However, for the driver to be used by the system, it has to
+ * be registered with the oFono core using ofono_sms_register().
+ *
+ * This is done once the modem driver determines that SMS is properly
+ * supported by the hardware.
+ */
 struct ofono_sms *ofono_sms_create(struct ofono_modem *modem,
 					unsigned int vendor,
 					const char *driver,
@@ -969,6 +987,14 @@ static void sms_load_settings(struct ofono_sms *sms, const char *imsi)
 
 }
 
+
+/*
+ * Indicate oFono that a SMS driver is ready for operation
+ *
+ * This is called after ofono_sms_create() was done and the modem
+ * driver determined that a modem supports SMS correctly. Once this
+ * call succeeds, the D-BUS interface for SMS goes live.
+ */
 void ofono_sms_register(struct ofono_sms *sms)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
@@ -978,16 +1004,16 @@ void ofono_sms_register(struct ofono_sms *sms)
 	struct ofono_atom *sim_atom;
 
 	if (!g_dbus_register_interface(conn, path,
-					SMS_MANAGER_INTERFACE,
+					OFONO_SMS_MANAGER_INTERFACE,
 					sms_manager_methods,
 					sms_manager_signals,
 					NULL, sms, NULL)) {
 		ofono_error("Could not create %s interface",
-				SMS_MANAGER_INTERFACE);
+				OFONO_SMS_MANAGER_INTERFACE);
 		return;
 	}
 
-	ofono_modem_add_interface(modem, SMS_MANAGER_INTERFACE);
+	ofono_modem_add_interface(modem, OFONO_SMS_MANAGER_INTERFACE);
 
 	sms->mw_watch = __ofono_modem_add_atom_watch(modem,
 					OFONO_ATOM_TYPE_MESSAGE_WAITING,
