@@ -110,6 +110,7 @@ enum stk_data_object_type {
 	STK_DATA_OBJECT_TYPE_CAUSE =				0x1A,
 	STK_DATA_OBJECT_TYPE_LOCATION_STATUS =			0x1B,
 	STK_DATA_OBJECT_TYPE_TRANSACTION_ID =			0x1C,
+	STK_DATA_OBJECT_TYPE_BCCH_CHANNEL_LIST =		0x1D,
 	STK_DATA_OBJECT_TYPE_ICON_ID =				0x1E,
 	STK_DATA_OBJECT_TYPE_ITEM_ICON_ID_LIST =		0x1F,
 	STK_DATA_OBJECT_TYPE_CARD_READER_STATUS =		0x20,
@@ -131,7 +132,7 @@ enum stk_data_object_type {
 	STK_DATA_OBJECT_TYPE_BROWSER_ID =			0x30,
 	STK_DATA_OBJECT_TYPE_URL =				0x31,
 	STK_DATA_OBJECT_TYPE_BEARER =				0x32,
-	STK_DATA_OBJECT_TYPE_PROVISIONING_FILE_REFERENCE =	0x33,
+	STK_DATA_OBJECT_TYPE_PROVISIONING_FILE_REF =		0x33,
 	STK_DATA_OBJECT_TYPE_BROWSER_TERMINATION_CAUSE =	0x34,
 	STK_DATA_OBJECT_TYPE_BEARER_DESCRIPTION =		0x35,
 	STK_DATA_OBJECT_TYPE_CHANNEL_DATA =			0x36,
@@ -544,6 +545,8 @@ struct stk_location_info {
 	unsigned short ci;
 	ofono_bool_t has_ext_ci;
 	unsigned short ext_ci;
+	ofono_bool_t has_eutran_ci;
+	guint32 eutran_ci;
 };
 
 /*
@@ -640,6 +643,7 @@ struct stk_r_apdu {
 
 /* Defined in TS 102.223 Section 8.38 */
 struct stk_timer_value {
+	ofono_bool_t has_value;
 	unsigned char hour;
 	unsigned char minute;
 	unsigned char second;
@@ -992,6 +996,21 @@ struct stk_command_language_notification {
 	char language[3];
 };
 
+struct stk_command_launch_browser {
+	unsigned char browser_id;
+	char *url;
+	struct stk_common_byte_array bearer;
+	GSList *prov_file_refs;
+	char *text_gateway_proxy_id;
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+	struct stk_common_byte_array network_name;
+	char *text_usr;
+	char *text_passwd;
+};
+
 struct stk_command {
 	unsigned char number;
 	unsigned char type;
@@ -1017,6 +1036,7 @@ struct stk_command {
 		struct stk_command_run_at_command run_at_command;
 		struct stk_command_send_dtmf send_dtmf;
 		struct stk_command_language_notification language_notification;
+		struct stk_command_launch_browser launch_browser;
 	};
 
 	void (*destructor)(struct stk_command *command);
@@ -1053,6 +1073,72 @@ struct stk_response_select_item {
 	unsigned char item_id;
 };
 
+struct stk_response_set_up_call {
+	struct stk_common_byte_array cc_requested_action;
+	struct {
+		ofono_bool_t cc_modified;
+		struct stk_result result;
+	} modified_result;
+};
+
+struct stk_response_local_info {
+	union {
+		struct stk_location_info location;
+		const char *imei;
+		struct stk_network_measurement_results {
+			struct stk_common_byte_array nmr;
+			struct stk_bcch_ch_list {
+				const short *channels;
+				int length;
+			} bcch_ch_list;
+		} nmr;
+		struct sms_scts datetime;
+		const char *language;
+		enum stk_battery_state battery_charge;
+		enum stk_access_technology_type access_technology;
+		struct stk_timing_advance {
+			enum {
+				STK_TIMING_ADVANCE_ME_IDLE = 0x00,
+				STK_TIMING_ADVANCE_ME_NOT_IDLE = 0x01,
+			} status;
+			/* Contains bit periods number according to 3GPP TS
+			 * 44.118 Section 9.3.106 / 3GPP TS 44.018 Section
+			 * 10.5.2.40.1, not microseconds */
+			int advance;
+		} tadv;
+		/* Bits[31:24]: manufacturer, bits[23:0]: serial number */
+		guint32 esn;
+		const char *imeisv;
+		enum stk_network_search_mode {
+			STK_NETWORK_SEARCH_MODE_MANUAL = 0x00,
+			STK_NETWORK_SEARCH_MODE_AUTOMATIC = 0x01,
+		} search_mode;
+		const char *meid;
+		struct stk_broadcast_network_information broadcast_network_info;
+		struct stk_access_technologies {
+			const enum stk_access_technology_type *techs;
+			int length;
+		} access_technologies;
+		struct {
+			struct stk_access_technologies access_techs;
+			struct stk_location_info *locations;
+		} location_infos;
+		struct {
+			struct stk_access_technologies access_techs;
+			struct stk_network_measurement_results *nmrs;
+		} nmrs;
+	};
+};
+
+struct stk_response_timer_mgmt {
+	unsigned char id;
+	struct stk_timer_value value;
+};
+
+struct stk_response_run_at_command {
+	const char *at_response;
+};
+
 struct stk_response {
 	unsigned char number;
 	unsigned char type;
@@ -1071,6 +1157,16 @@ struct stk_response {
 		struct stk_response_generic set_up_menu;
 		struct stk_response_select_item select_item;
 		struct stk_response_generic send_sms;
+		struct stk_response_set_up_call set_up_call;
+		struct stk_response_generic polling_off;
+		struct stk_response_local_info provide_local_info;
+		struct stk_response_generic set_up_event_list;
+		struct stk_response_timer_mgmt timer_mgmt;
+		struct stk_response_generic set_up_idle_mode_text;
+		struct stk_response_run_at_command run_at_command;
+		struct stk_response_generic send_dtmf;
+		struct stk_response_generic language_notification;
+		struct stk_response_generic launch_browser;
 	};
 
 	void (*destructor)(struct stk_response *response);
